#include "piece_logic.h"
#include "king.h"
#include "pawn.h"
#include <stdint.h>
#include <sys/types.h>


/**
 * @brief Function for determining if the piece is on the white team or on the black team.
 * @note   Pieces should be enumerated even if white and odd if black.
 * @param  piece: The piece under question.
 * @retval Return true if on the white team, else false.
 */
bool white_team(uint8_t piece)
{
    return ((piece % 2u) == 0u);
}

/**
 * @brief  Function for determining if two pieces are on the same team or not.
 * @param  piece_one: Piece one ofcoarse.
 * @param  piece_two: Piece two obviously.
 * @retval True if on opposite teams, else false.
 */
bool opposite_teams(uint8_t piece_one, uint8_t piece_two)
{
    return (((piece_one % 2u) == 0u) ^ ((piece_two % 2u) == 0u));
}

/**
 * @brief  Check to see if the square is safe from the other team.
 * @param  column: Column of potential move
 * @param  row: Row of the potential move
 * @retval True if the square is safe, else is false
 */
bool square_is_safe(uint8_t row, uint8_t column, Game_State_t * game_state)
{
    int temp = (int)row + (game_state->player_turn ? -1 : 1);

    /* first check if pawns can take us */
    if ((column > 0) && ((game_state->player_turn ? PAWN_BLACK : PAWN_WHITE) == game_state->board_pieces[temp*8+(int)column - 1]))
    {
        //can be eaten by a pawn, not safe.
        return false;
    }
    if ((column < 7) && ((game_state->player_turn ? PAWN_BLACK : PAWN_WHITE) == game_state->board_pieces[temp*8+(int)column + 1]))
    {
        //can be eaten by a pawn, not safe.
        return false;
    }

    /* Other King */
    int start_r = (row == 0u) ? 0 : -1;
    int stop_r = (row == 7u) ? 0 : 1;
    int start_c = (column == 0u) ? 0 : -1;
    int stop_c = (column == 7u) ? 0 : 1;
    for (int up_down = start_r; up_down <= stop_r; up_down++)
    {
        for (int left_right = start_c; left_right <= stop_c; left_right++)
        {
            int x = row + left_right;
            int y = column + up_down;
            if ((game_state->player_turn ? KING_BLACK : KING_WHITE) == game_state->board_pieces[x*8+y])
            {
                return false;
            }
        }
    }

    for (uint8_t direction = 0u; direction < 4u; direction++)
    {
        int8_t up_down_step;
        int8_t left_right_step;
        if (direction == 0u)
        {
            up_down_step = 1;
            left_right_step = 0;
        }
        else if (direction == 1u)
        {
            up_down_step = 0;
            left_right_step = 1;
        }
        else if (direction == 2u)
        {
            up_down_step = -1;
            left_right_step = 0;
        }
        else
        {
            up_down_step = 0;
            left_right_step = -1;
        }
        /* Rooks and queens */
        bool loop = true;
        int x = (int)row;
        int y = (int)column;
        while (loop)
        {
            x += left_right_step;
            y += up_down_step;
            if ((x < 0) || (y < 0) || (x >= 8) || (y >= 8))
            {
                loop = false;
            }
            else if (game_state->board_pieces[x*8+y] == SQUARE_EMPTY)
            {
                /* Do nothing */
            }
            else if (((game_state->player_turn ? ROOK_BLACK : ROOK_WHITE) == game_state->board_pieces[x*8+y]) || ((game_state->player_turn ? QUEEN_BLACK : QUEEN_WHITE) == game_state->board_pieces[x*8+y]))
            {
                return false;
            }
            else
            {
                loop = false;
            }
        }
    }
    /* Bishops and queens */
    for (uint8_t direction = 0u; direction < 4u; direction++)
    {
        int8_t up_down_step;
        int8_t left_right_step;
        if (direction == 0u)
        {
            up_down_step = 1;
            left_right_step = 1;
        }
        else if (direction == 1u)
        {
            up_down_step = -1;
            left_right_step = 1;
        }
        else if (direction == 2u)
        {
            up_down_step = -1;
            left_right_step = -1;
        }
        else
        {
            up_down_step = 1;
            left_right_step = -1;
        }
        bool loop = true;
        int x = (int)row;
        int y = (int)column;
        while (loop)
        {
            uint8_t bish = (game_state->player_turn ? BISHOP_BLACK : BISHOP_WHITE);
            uint8_t queen = (game_state->player_turn ? QUEEN_BLACK : QUEEN_WHITE);
            x += left_right_step;
            y += up_down_step;
            if ((x < 0) || (y < 0) || (x >= 8) || (y >= 8))
            {
                loop = false;
            }
            else if (game_state->board_pieces[x*8+y] == SQUARE_EMPTY)
            {
                /* do nothing */
            }
            else if ((bish == game_state->board_pieces[x*8+y]) || (queen == game_state->board_pieces[x*8+y]))
            {
                return false;
            }
            else
            {
                loop = false;
            }
        }
    }
    /* Knights */
    for (uint8_t direction = 0u; direction < 4u; direction++)
    {
        int8_t up_down_step;
        int8_t left_right_step;
        if (direction == 0u)
        {
            up_down_step = 2;
            left_right_step = 0;
        }
        else if (direction == 1u)
        {
            up_down_step = 0;
            left_right_step = 2;
        }
        else if (direction == 2u)
        {
            up_down_step = -2;
            left_right_step = 0;
        }
        else
        {
            up_down_step = 0;
            left_right_step = -2;
        }

        for (uint8_t i = 0u; i < 2u; i++)
        {
            if ((direction % 2u) == 0u)
            {
                left_right_step = (i == 0u) ? -1 : 1;
            }
            else
            {
                up_down_step = (i == 0u) ? -1 : 1;
            }

            int8_t x = (int8_t)row + left_right_step;
            int8_t y = (int8_t)column + up_down_step;
            if ((x >= 0) && (y >= 0) && (x < 8) && (y < 8))
            {
                if ((game_state->player_turn ? KNIGHT_BLACK : KNIGHT_WHITE) == game_state->board_pieces[x*8+y])
                {
                    return false;
                }
            }
        }
    }
    return true;
}


/**
 * @brief  Function for "casting" a ray in any direction, vertical, horizontal, or diagonal. If the ray hits someone from the other team
 *         or the end of the board the array will be terminated.
 * @param  direction_r: Row direction
 * @param  direction_c: Column direction
 * @param  row: current row location
 * @param  column: current column location
 * @param  piece: the piece that is casting the ray.
 * @retval None
 */
bool cast_a_ray(uint8_t piece, int8_t direction_r, int8_t direction_c, uint8_t column, uint8_t row, Game_State_t * game_state)
{
    bool ret_val = false;
    bool loop = true;
    int x = row;
    int y = column;
    while (loop)
    {
        x += direction_c;
        y += direction_r;
        if((x < 0) || (y < 0) || (x >= 8) || (y >= 8))
        {
            loop = false;
        }
        else if (game_state->board_pieces[x*8+y] == SQUARE_EMPTY)
        {
            ret_val = Set_Light(piece, (uint8_t)x, (uint8_t)y, POTENTIAL_MOVE) || ret_val;
        }
        else if (opposite_teams(piece, game_state->board_pieces[x*8+y]))
        {
            ret_val = Set_Light(piece, (uint8_t)x, (uint8_t)y, POTENTIAL_TAKE) || ret_val;
            /* once we take a piece we can no longer take anymore */
            loop = false;
        }
        else
        {
            loop = false;
        }
    }
    return ret_val;
}


bool Check_If_Player_Can_Move(bool white, Game_State_t * game_state)
{
    for (uint8_t row = 0; row < 8u; row++)
    {
        for (uint8_t column = 0; column < 8u; column++)
        {
            if((white_team(game_state->board_pieces[row*8+column]) == white))
            {
                // SDL_Log("move: Row:%d, Col:%d", row, column);

                if(Mark_Potential_Moves(game_state->board_pieces[row*8+column], column, row, game_state))
                {
                    // SDL_Log("Player Can still move: Row:%d, Col:%d", row, column);
                    clear_board_state();
                    return true;
                }
            }
        }

    }
    clear_board_state();
    // SDL_Log("Player cant move");
    return false;
}


/**
 * @brief  Function for marking the potential moves.
 * @param  piece: Piece that we are marking the potential moves for.
 * @param  row: Current row location of the piece.
 * @param  column: Current column location of the piece.
 * @retval None
 */
bool Mark_Potential_Moves(uint8_t piece, uint8_t column, uint8_t row, Game_State_t * game_state)
{
    bool ret_val = false;
    switch (piece)
    {
        case PAWN_WHITE :
        case PAWN_BLACK:
        {
            int direction = white_team(piece) ? -1 : 1;
            uint8_t temp_row = (uint8_t)((int)row + direction);
            if (row == (white_team(piece) ? 6u : 1u))
            {
                if(game_state->board_pieces[temp_row*8u+column] == SQUARE_EMPTY)
                {
                    ret_val = pawn_move(piece, (uint8_t)((int)row + (direction * 2)), column, game_state) || ret_val;
                }
            }

            ret_val = pawn_move(piece, temp_row, column, game_state) || ret_val;
            if(column != 0)
            {
                ret_val = pawn_take(piece, temp_row, column - 1u, game_state) || ret_val;
            }
            if(column != 7u)
            {
                ret_val = pawn_take(piece, temp_row, column + 1u, game_state) || ret_val;
            }

            break;
        }
        case ROOK_WHITE:
        case ROOK_BLACK:
        {
            for (uint8_t direction = 0u; direction < 4u; direction++)
            {
                int8_t up_down_step;
                int8_t left_right_step;
                if(direction == 0u)
                {
                    up_down_step = 1;
                    left_right_step = 0;
                }
                else if(direction == 1u)
                {
                    up_down_step = 0;
                    left_right_step = 1;
                }
                else if (direction == 2u)
                {
                    up_down_step = -1;
                    left_right_step = 0;
                }
                else
                {
                    up_down_step = 0;
                    left_right_step = -1;
                }
                ret_val = cast_a_ray(piece, up_down_step, left_right_step, column, row, game_state) || ret_val;
            }
            break;

        }
        case KNIGHT_WHITE:
        case KNIGHT_BLACK:
        {
            for (uint8_t direction = 0u; direction < 4u; direction++)
            {
                int8_t up_down_step;
                int8_t left_right_step;
                if (direction == 0u)
                {
                    up_down_step = 2;
                    left_right_step = 0;
                }
                else if (direction == 1u)
                {
                    up_down_step = 0;
                    left_right_step = 2;
                }
                else if (direction == 2u)
                {
                    up_down_step = -2;
                    left_right_step = 0;
                }
                else
                {
                    up_down_step = 0;
                    left_right_step = -2;
                }

                for (uint8_t i = 0u; i < 2u; i++)
                {
                    if((direction % 2u) == 0u)
                    {
                        left_right_step = (i == 0u) ? -1 : 1;
                    }
                    else
                    {
                        up_down_step = (i == 0u) ? -1 : 1;
                    }

                    int8_t x = (int8_t)row + left_right_step;
                    int8_t y = (int8_t)column + up_down_step;
                    if ((x >= 0) && (y >= 0) && (x < 8) && (y < 8))
                    {
                        if (game_state->board_pieces[x*8+y] == SQUARE_EMPTY)
                        {
                            ret_val = Set_Light(piece, (uint8_t)x, (uint8_t)y, POTENTIAL_MOVE) || ret_val;
                        }
                        else if (opposite_teams(piece, game_state->board_pieces[x*8+y]))
                        {
                            ret_val = Set_Light(piece, (uint8_t)x, (uint8_t)y, POTENTIAL_TAKE) || ret_val;
                        }
                    }
                }
            }
            break;
        }
        case BISHOP_WHITE:
        case BISHOP_BLACK:
        {
            for (uint8_t direction = 0u; direction < 4u; direction++)
            {
                int8_t up_down_step;
                int8_t left_right_step;
                if (direction == 0u)
                {
                    up_down_step = 1;
                    left_right_step = 1;
                }
                else if (direction == 1u)
                {
                    up_down_step = -1;
                    left_right_step = 1;
                }
                else if (direction == 2u)
                {
                    up_down_step = -1;
                    left_right_step = -1;
                }
                else
                {
                    up_down_step = 1;
                    left_right_step = -1;
                }
                ret_val = cast_a_ray(piece, up_down_step, left_right_step, column, row, game_state) || ret_val;
            }
            break;
        }
        case QUEEN_WHITE:
        case QUEEN_BLACK:
        {
            //Mark bishop moves
            ret_val = Mark_Potential_Moves((piece - 2u), column, row, game_state) || ret_val;
            //Mark rook moves
            ret_val = Mark_Potential_Moves((piece - 6u), column, row, game_state) || ret_val;
            break;
        }
        case KING_WHITE:
        case KING_BLACK:
        {
            int8_t start_r = (row == 0u) ? 0 : -1;
            int8_t stop_r = (row == 7u) ? 0 : 1;
            int8_t start_c = (column == 0u) ? 0 : -1;
            int8_t stop_c = (column == 7u) ? 0 : 1;
            for (int8_t left_right = start_r; left_right <= stop_r; left_right++)
            {
                for (int8_t up_down = start_c; up_down <= stop_c; up_down++)
                {
                    int8_t x = row + left_right;
                    int8_t y = column + up_down;
                    if (square_is_safe(x, y, game_state))
                    {
                        if (game_state->board_pieces[x*8+y] == SQUARE_EMPTY)
                        {
                            game_state->board_state[x*8+y] = POTENTIAL_MOVE;
                            ret_val = true;
                        }
                        else if (opposite_teams(piece, game_state->board_pieces[x*8+y]))
                        {
                            game_state->board_state[x*8+y] = POTENTIAL_TAKE;
                            ret_val = true;
                        }
                    }
                }
            }
            uint8_t white_black_idx = white_team(piece) ? 0u : 1u;
            uint8_t kings_row = white_team(piece) ? 7u : 0u;

            //Can only castle if not currently in check
            if (square_is_safe(row, column, game_state))
            {
                // Queen side castle
                if(game_state->castling_allowed[white_black_idx][1u] && (game_state->board_pieces[kings_row*8+1u] == SQUARE_EMPTY)
                && (game_state->board_pieces[kings_row*8+2u] == SQUARE_EMPTY) && (game_state->board_pieces[kings_row*8+3u]) == SQUARE_EMPTY)
                {
                    //First Check to see if the king will pass through check
                    if(square_is_safe(kings_row, 3u, game_state) && square_is_safe(kings_row, 2u, game_state))
                    {
                        // Yay we can castle queen side!
                        game_state->board_state[kings_row*8+2u] = POTENTIAL_CASTLE;
                        ret_val = true;
                    }
                }

                // King side castle
                if (game_state->castling_allowed[white_black_idx][0u] && (game_state->board_pieces[kings_row*8+5u] == SQUARE_EMPTY) && (game_state->board_pieces[kings_row*8+6u] == SQUARE_EMPTY))
                {
                    //First Check to see if the king will pass through check
                    if(square_is_safe(kings_row, 5u, game_state) && square_is_safe(kings_row, 6u, game_state))
                    {
                        // Yay we can castle king side!
                        game_state->board_state[kings_row*8+6u] = POTENTIAL_CASTLE;
                        ret_val = true;
                    }
                }
            }

            break;
        }

        default:
            break;
    }
    return ret_val;
}

